"""
Split a multiple-sequence FASTA file into many single-sequence FASTA files
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import os

import utils


def parseArgs():
    description = ("Split a multiple-sequence FASTA file into many "
                   "single-sequence FASTA files")
    parser = argparse.ArgumentParser(
        description=description)
    parser.add_argument("fastaFile", help="the FASTA file to split")
    args = parser.parse_args()
    return args


def decompressFasta(args):
    fastaFileName = args.fastaFile
    filename, extension = os.path.splitext(fastaFileName)
    if extension == '.gz':
        utils.log("Decompressing {}".format(fastaFileName))
        cmd = "gunzip {}".format(fastaFileName)
        utils.runCommand(cmd)
        fastaFileName = filename
    return fastaFileName


def splitFasta(fastaFileName):
    splitFileNames = []
    with open(fastaFileName) as fastaFile:
        currentFile = None
        currentFileName = None
        for line in fastaFile:
            if line[0] == '>':
                if currentFile is not None:
                    currentFile.close()
                currentFileName = line[1:].split()[0].strip() + '.fa'
                utils.log("Creating {}".format(currentFileName))
                splitFileNames.append(currentFileName)
                currentFile = open(currentFileName, 'w')
            currentFile.write(line)
        currentFile.close()
    return splitFileNames


def compressSplits(splitFileNames):
    compressedFileNames = []
    for splitFileName in splitFileNames:
        utils.log("Compressing {}".format(splitFileName))
        cmd = "bgzip {}".format(splitFileName)
        utils.runCommand(cmd)
        compressedFileName = "{}.gz".format(splitFileName)
        compressedFileNames.append(compressedFileName)
    return compressedFileNames


def indexSplits(compressedFileNames):
    for compressedFileName in compressedFileNames:
        utils.log("Indexing {}".format(compressedFileName))
        cmd = "samtools faidx {}".format(compressedFileName)
        utils.runCommand(cmd)


@utils.Timed()
def main():
    requiredExecutables = ['gunzip', 'bgzip', 'samtools']
    utils.requireExecutables(requiredExecutables)
    args = parseArgs()
    fastaFileName = decompressFasta(args)
    splitFileNames = splitFasta(fastaFileName)
    compressedFileNames = compressSplits(splitFileNames)
    indexSplits(compressedFileNames)


if __name__ == '__main__':
    main()



from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.protocol as protocol


def getValueListName(cls):
    return [field for field in cls.DESCRIPTOR.fields_by_name
            if field != 'next_page_token'][0]


class SearchResponseBuilderTest(unittest.TestCase):
    """
    Tests the SearchResponseBuilder class to ensure that it behaves
    correctly.
    """

    def testIntegrity(self):
        # Verifies that the values we put in are exactly what we get
        # back across all subclasses of SearchResponse
        for class_ in [responseClass for _, _, responseClass in
                       protocol.postMethods]:
            instance = class_()
            valueList = getattr(instance, getValueListName(class_))
            valueList.add()
            builder = protocol.SearchResponseBuilder(
                class_, len(valueList), 2 ** 32)
            for value in valueList:
                builder.addValue(value)
            builder.setNextPageToken(instance.next_page_token)
            otherInstance = protocol.fromJson(
                builder.getSerializedResponse(), class_)
            self.assertEqual(instance, otherInstance)

    def testPageSizeOverflow(self):
        # Verifies that the page size behaviour is correct when we keep
        # filling after full is True.
        responseClass = protocol.SearchVariantsResponse
        valueClass = protocol.Variant
        for pageSize in range(1, 10):
            builder = protocol.SearchResponseBuilder(
                responseClass, pageSize, 2 ** 32)
            self.assertEqual(builder.getPageSize(), pageSize)
            self.assertFalse(builder.isFull())
            for listLength in range(1, 2 * pageSize):
                builder.addValue(valueClass())
                instance = protocol.fromJson(
                    builder.getSerializedResponse(), responseClass)

                valueList = getattr(instance, getValueListName(responseClass))
                self.assertEqual(len(valueList), listLength)
                if listLength < pageSize:
                    self.assertFalse(builder.isFull())
                else:
                    self.assertTrue(builder.isFull())

    def testPageSizeExactFill(self):
        responseClass = protocol.SearchVariantsResponse
        valueClass = protocol.Variant
        for pageSize in range(1, 10):
            builder = protocol.SearchResponseBuilder(
                responseClass, pageSize, 2 ** 32)
            self.assertEqual(builder.getPageSize(), pageSize)
            while not builder.isFull():
                builder.addValue(valueClass())
            instance = protocol.fromJson(builder.getSerializedResponse(),
                                         responseClass)
            valueList = getattr(instance, getValueListName(responseClass))
            self.assertEqual(len(valueList), pageSize)

    def testMaxBufferSizeOverridesPageSize(self):
        responseClass = protocol.SearchVariantsResponse
        typicalValue = protocol.Variant()
        # We have to put some values in here or it will have zero length.
        typicalValue.start = 1
        typicalValue.end = 2
        typicalValue.reference_bases = "AAAAAAAA"
        typicalValueLength = typicalValue.ByteSize()
        for numValues in range(1, 10):
            maxBufferSize = numValues * typicalValueLength
            builder = protocol.SearchResponseBuilder(
                responseClass, 1000, maxBufferSize)
            self.assertEqual(
                maxBufferSize, builder.getMaxBufferSize())
            while not builder.isFull():
                builder.addValue(typicalValue)
            instance = protocol.fromJson(builder.getSerializedResponse(),
                                         responseClass)
            valueList = getattr(instance, getValueListName(responseClass))
            self.assertEqual(len(valueList), numValues)

    def testNextPageToken(self):
        responseClass = protocol.SearchVariantsResponse
        builder = protocol.SearchResponseBuilder(
            responseClass, 100, 2 ** 32)
        # If not set, pageToken should be empty string
        self.assertIsNone(builder.getNextPageToken())
        instance = protocol.fromJson(builder.getSerializedResponse(),
                                     responseClass)
        self.assertEqual(instance.next_page_token, "")
        # page tokens can be any string.
        for nextPageToken in ["", "string"]:
            builder.setNextPageToken(nextPageToken)
            self.assertEqual(nextPageToken, builder.getNextPageToken())
            instance = protocol.fromJson(builder.getSerializedResponse(),
                                         responseClass)
            self.assertEqual(nextPageToken, instance.next_page_token)



"""
Tests the data interfacing layer for various objects
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.client as client
import ga4gh.datarepo as datarepo
import ga4gh.backend as backend
import tests.paths as paths
import tests.utils as utils


class TestInterfacingLayer(unittest.TestCase):
    """
    Test fetching objects via the db and via the client return
    identical results
    """
    def _assertEqual(self, dbObj, pbObj):
        dbToPb = dbObj.toProtocolElement()
        self.assertEqual(dbToPb, pbObj)

    def _getPageSizes(self, num):
        if num <= 0:
            return []
        return [None, 1, 2, 3, num - 1, num, num + 1]

    def _testGetMethod(self, repoIteratorMethod, clientGetMethod):
        repoObjs = repoIteratorMethod()
        for repoObj in repoObjs:
            id_ = repoObj.getId()
            clientObj = clientGetMethod(id_)
            self._assertEqual(repoObj, clientObj)

    def _testSearchMethod(self, repoIteratorMethod, clientSearchMethod):
        repoObjs = list(repoIteratorMethod())
        numRepoObjs = len(repoObjs)
        pageSizes = self._getPageSizes(numRepoObjs)
        for pageSize in pageSizes:
            self._client.set_page_size(pageSize)
            clientObjs = list(clientSearchMethod())
            for repoObj, clientObj in utils.zipLists(repoObjs, clientObjs):
                self._assertEqual(repoObj, clientObj)

    def _testSearchMethodInContainer(
            self, containerAccessorMethod, clientSearchMethod,
            containerIteratorMethod, equalMethod="_assertEqual"):
        for container in containerIteratorMethod:
            containerId = container.getId()
            repoObjs = getattr(container, containerAccessorMethod)()
            numRepoObjs = len(repoObjs)
            pageSizes = self._getPageSizes(numRepoObjs)
            for pageSize in pageSizes:
                self._client.set_page_size(pageSize)
                clientObjs = list(clientSearchMethod(containerId))
                for repoObj, clientObj in utils.zipLists(
                        repoObjs, clientObjs):
                    assertEqual = getattr(self, equalMethod)
                    assertEqual(repoObj, clientObj)

    def setUp(self):
        self._repo = datarepo.SqlDataRepository(paths.testDataRepo)
        self._repo.open(datarepo.MODE_READ)
        self._backend = backend.Backend(self._repo)
        self._client = client.LocalClient(self._backend)

    def testGetBioSample(self):
        self._testGetMethod(
            self._repo.allBioSamples, self._client.get_bio_sample)

    def testGetIndividual(self):
        self._testGetMethod(
            self._repo.allIndividuals, self._client.get_individual)

    def testGetDataset(self):
        self._testGetMethod(
            self._repo.getDatasets, self._client.get_dataset)

    def testGetReadGroupSet(self):
        self._testGetMethod(
            self._repo.allReadGroupSets, self._client.get_read_group_set)

    def testGetReadGroup(self):
        self._testGetMethod(
            self._repo.allReadGroups, self._client.get_read_group)

    def testGetCallSet(self):
        self._testGetMethod(
            self._repo.allCallSets, self._client.get_call_set)

    def testGetVariant(self):
        repoVariantSets = self._repo.allVariantSets()
        referenceName = self._repo.getReferenceSets(
            )[0].getReferences()[0].getName()
        for variantSet in repoVariantSets:
            repoVariants = variantSet.getVariants(referenceName, 0, 2**32)
            for repoVariant in repoVariants:
                variantId = repoVariant.getId()
                variant = self._client.get_variant(variantId)
                self._assertEqual(repoVariant, variant)

    def testGetVariantSet(self):
        self._testGetMethod(
            self._repo.allVariantSets, self._client.get_variant_set)

    def testGetVariantAnnotationSet(self):
        self._testGetMethod(
            self._repo.allVariantAnnotationSets,
            self._client.get_variant_annotation_set)

    def testGetFeatureSet(self):
        self._testGetMethod(
            self._repo.allFeatureSets, self._client.get_feature_set)

    def testGetFeature(self):
        repoFeatures = self._repo.allFeatures()
        for repoFeature in repoFeatures:
            featureId = repoFeature.id
            feature = self._client.get_feature(featureId)
            self.assertEqual(repoFeature, feature)

    def testGetReferenceSet(self):
        self._testGetMethod(
            self._repo.getReferenceSets, self._client.get_reference_set)

    def testGetReference(self):
        self._testGetMethod(
            self._repo.allReferences, self._client.get_reference)

    def testGetRnaQuantificationSet(self):
        self._testGetMethod(
            self._repo.allRnaQuantificationSets,
            self._client.get_rna_quantification_set)

    def testGetRnaQuantification(self):
        self._testGetMethod(
            self._repo.allRnaQuantifications,
            self._client.get_rna_quantification)

    def testGetExpressionLevel(self):
        self._testGetMethod(
            self._repo.allExpressionLevels,
            self._client.get_expression_level)

    def testSearchVariants(self):
        for variantSet in self._repo.allVariantSets():
            referenceSet = variantSet.getReferenceSet()
            referenceName = referenceSet.getLocalId()
            variantSetId = variantSet.getId()
            repoVariants = list(
                variantSet.getVariants(referenceName, 0, 2**32))
            numRepoVariants = len(repoVariants)
            pageSizes = self._getPageSizes(numRepoVariants)
            for pageSize in pageSizes:
                self._client.set_page_size(pageSize)
                variants = self._client.search_variants(variantSetId)
                for repoVariant, variant in utils.zipLists(
                        repoVariants, variants):
                    self._assertEqual(repoVariant, variant)

    def testSearchVariantAnnotations(self):
        for variantAnnotationSet in self._repo.allVariantAnnotationSets():
            variantAnnotationSetId = variantAnnotationSet.getId()
            variantSet = variantAnnotationSet._parentContainer
            referenceSet = variantSet.getReferenceSet()
            referenceName = referenceSet.getLocalId()
            repoVariantAnnotations = list(
                variantAnnotationSet.getVariantAnnotations(
                    referenceName, 0, 2**32))
            numRepoVariantAnnotations = len(repoVariantAnnotations)
            pageSizes = self._getPageSizes(numRepoVariantAnnotations)
            for pageSize in pageSizes:
                self._client.set_page_size(pageSize)
                variantAnnotations = self._client.search_variant_annotations(
                    variantAnnotationSetId, referenceName=referenceName)
                for repoVa, va in utils.zipLists(
                        repoVariantAnnotations, variantAnnotations):
                    self._assertEqual(repoVa, va)

    def testSearchFeatures(self):
        repoFeatureSets = self._repo.allFeatureSets()
        for repoFeatureSet in repoFeatureSets:
            repoFeatures = list(repoFeatureSet.getFeatures())
            numFeatures = len(repoFeatures)
            featureSetId = repoFeatureSet.getId()
            pageSizes = self._getPageSizes(numFeatures)
            for pageSize in pageSizes:
                features = list(self._client.search_features(featureSetId))
                for repoFeature, feature in utils.zipLists(
                        repoFeatures, features):
                    self.assertEqual(repoFeature, feature)

    def testSearchDatasets(self):
        self._testSearchMethod(
            self._repo.getDatasets, self._client.search_datasets)

    def testSearchVariantSets(self):
        self._testSearchMethodInContainer(
            'getVariantSets',
            self._client.search_variant_sets,
            self._repo.getDatasets())

    def testSearchVariantAnnotationSets(self):
        self._testSearchMethodInContainer(
            'getVariantAnnotationSets',
            self._client.search_variant_annotation_sets,
            self._repo.allVariantSets())

    def testSearchFeatureSets(self):
        self._testSearchMethodInContainer(
            'getFeatureSets',
            self._client.search_feature_sets,
            self._repo.getDatasets())

    def testSearchCallSets(self):
        self._testSearchMethodInContainer(
            'getCallSets',
            self._client.search_call_sets,
            self._repo.allVariantSets())

    def testSearchBioSamples(self):
        self._testSearchMethodInContainer(
            'getBioSamples',
            self._client.search_bio_samples,
            self._repo.getDatasets())

    def testSearchIndividuals(self):
        self._testSearchMethodInContainer(
            'getIndividuals',
            self._client.search_individuals,
            self._repo.getDatasets())

    def testSearchReadGroupSets(self):
        self._testSearchMethodInContainer(
            'getReadGroupSets',
            self._client.search_read_group_sets,
            self._repo.getDatasets())

    def testSearchReads(self):
        for repoReadGroup in self._repo.allReadGroups():
            repoReferenceSet = \
                repoReadGroup._parentContainer.getReferenceSet()
            for repoReference in repoReferenceSet.getReferences():
                referenceId = repoReference.getId()
                readGroupId = repoReadGroup.getId()
                repoReads = list(
                    repoReadGroup.getReadAlignments(repoReference))
                numRepoReads = len(repoReads)
                pageSizes = self._getPageSizes(numRepoReads)
                for pageSize in pageSizes:
                    self._client.set_page_size(pageSize)
                    reads = list(self._client.search_reads(
                        [readGroupId], referenceId))
                    for repoRead, read in utils.zipLists(
                            repoReads, reads):
                        self.assertEqual(repoRead, read)

    def testSearchReferenceSets(self):
        self._testSearchMethod(
            self._repo.getReferenceSets, self._client.search_reference_sets)

    def testSearchReferences(self):
        repoReferenceSets = self._repo.getReferenceSets()
        for repoReferenceSet in repoReferenceSets:
            referenceSetId = repoReferenceSet.getId()
            repoReferences = repoReferenceSet.getReferences()
            numRefs = len(repoReferences)
            pageSizes = self._getPageSizes(numRefs)
            for pageSize in pageSizes:
                self._client.set_page_size(pageSize)
                references = list(
                    self._client.search_references(referenceSetId))
                for repoReference, reference in utils.zipLists(
                        repoReferences, references):
                    self._assertEqual(repoReference, reference)

    # search_phenotype does not return containers, it returns
    # PhenotypeInstances. So, we iterate through the container
    # (PhenotypeAssociationSet), and compare the leaf nodes (PhenotypeInstance)
    def testSearchPhenotypes(self):
        phenotypeAssociationSets = self._repo.allPhenotypeAssociationSets()
        for phenotypeAssociationSet in phenotypeAssociationSets:
            repoAssociations = list(phenotypeAssociationSet.getAssociations())
            numAssociations = len(repoAssociations)
            phenotypeAssociationSetId = phenotypeAssociationSet.getId()
            pageSizes = self._getPageSizes(numAssociations)
            for pageSize in pageSizes:
                clientPhenotypes = list(self._client.search_phenotype
                                        (phenotypeAssociationSetId))
                for repoAssociation, clientPhenotype in utils.zipLists(
                        repoAssociations, clientPhenotypes):
                    self.assertEqual(repoAssociation.phenotype,
                                     clientPhenotype)

    def testSearchPhenotypeAssociationSets(self):
        self._testSearchMethodInContainer(
            'getPhenotypeAssociationSets',
            self._client.search_phenotype_association_sets,
            self._repo.getDatasets())

    def testSearchGenotypePhenotypes(self):
        self._testSearchMethodInContainer(
            'getAssociations',
            self._client.search_genotype_phenotype,
            self._repo.allPhenotypeAssociationSets(),
            equalMethod='assertEqual')

    def testSearchRnaQuantificationSets(self):
        self._testSearchMethodInContainer(
            'getRnaQuantificationSets',
            self._client.search_rna_quantification_sets,
            self._repo.getDatasets())

    def testSearchRnaQuantifications(self):
        self._testSearchMethodInContainer(
            'getRnaQuantifications',
            self._client.search_rna_quantifications,
            self._repo.allRnaQuantificationSets())

    def testSearchExpressionLevels(self):
        self._testSearchMethodInContainer(
            'getExpressionLevels',
            self._client.search_expression_levels,
            self._repo.allRnaQuantifications())



"""
End-to-end tests for the simulator configuration. Sets up a server with
the backend, sends some basic queries to that server and verifies results
are as expected.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest
import logging
import random
import ga4gh.datamodel.reads as reads
import ga4gh.datamodel.references as references
import ga4gh.datamodel.variants as variants
import ga4gh.datamodel.sequence_annotations as sequence_annotations
import ga4gh.frontend as frontend
import ga4gh.protocol as protocol


class TestSimulatedStack(unittest.TestCase):
    """
    Tests the full stack for the Simulated backend by using the Flask
    testing client.
    """
    @classmethod
    def setUpClass(cls):
        # silence usually unhelpful CORS log
        logging.getLogger('ga4gh.frontend.cors').setLevel(logging.CRITICAL)
        # Set the random seed to make tests reproducible.
        random.seed(1)
        config = {
            "DATA_SOURCE": "simulated://",
            "SIMULATED_BACKEND_RANDOM_SEED": 1112,
            "SIMULATED_BACKEND_NUM_CALLS": 5,
            "SIMULATED_BACKEND_VARIANT_DENSITY": 1.0,
            "SIMULATED_BACKEND_NUM_VARIANT_SETS": 4,
            "SIMULATED_BACKEND_NUM_REFERENCE_SETS": 3,
            "SIMULATED_BACKEND_NUM_REFERENCES_PER_REFERENCE_SET": 4,
            "SIMULATED_BACKEND_NUM_ALIGNMENTS_PER_READ_GROUP": 5,
            # "DEBUG": True
        }
        frontend.reset()
        frontend.configure(
            baseConfig="TestConfig", extraConfig=config)
        cls.app = frontend.app.test_client()

    @classmethod
    def tearDownClass(cls):
        cls.app = None

    def setUp(self):
        self.backend = frontend.app.backend
        self.dataRepo = self.backend.getDataRepository()
        self.dataset = self.dataRepo.getDatasets()[0]
        self.readGroupSet = self.dataset.getReadGroupSets()[0]
        self.readGroup = self.readGroupSet.getReadGroups()[0]
        self.reference = \
            self.readGroupSet.getReferenceSet().getReferences()[0]
        self.variantSet = self.dataset.getVariantSets()[0]
        self.variantAnnotationSet = \
            self.variantSet.getVariantAnnotationSets()[0]
        self.backend.setMaxResponseLength(10000)

    def getBadIds(self):
        """
        Returns a list of IDs that should not exist in the server and should
        raise a 404 error.
        """
        return ["1234:", "x"*100, ":", ":xx", "::", ":::", "::::"]

    def sendJsonPostRequest(self, path, data):
        """
        Sends a JSON request to the specified path with the specified data
        and returns the response.
        """
        return self.app.post(
            path, headers={'Content-type': 'application/json'},
            data=data)

    def sendSearchRequest(self, path, request, responseClass):
        """
        Sends the specified protocol request instance as JSON, and
        parses the result into an instance of the specified response.
        """
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        self.assertEqual(200, response.status_code)
        responseData = protocol.fromJson(response.data, responseClass)
        self.assertTrue(
            protocol.validate(
                protocol.toJson(responseData),
                type(responseData)))
        return responseData

    def sendObjectGetRequest(self, path, id_):
        """
        Sends a GET request to the specified path for an object with the
        specified ID and returns the response.
        """
        return self.app.get("{}/{}".format(path, id_))

    def sendGetObject(self, path, id_, responseClass):
        """
        Sends a get request and parses the value into an instance of the
        specified class.
        """
        response = self.sendObjectGetRequest(path, id_)
        self.assertEqual(200, response.status_code)
        obj = protocol.fromJson(response.data, responseClass)
        self.assertIsInstance(obj, responseClass)
        return obj

    def sendListReferenceBasesRequest(self, request):
        """
        Sends a ListReferenceBasesRequest and parses the result into a
        ListReferenceBasesResponse.
        """
        path = '/listreferencebases'
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        self.assertEqual(response.status_code, 200)
        obj = protocol.fromJson(
            response.data, protocol.ListReferenceBasesResponse)
        self.assertIsInstance(obj, protocol.ListReferenceBasesResponse)
        return obj

    def verifyVariantSetsEqual(self, gaVariantSet, variantSet):
        dataset = variantSet.getParentContainer()
        self.assertEqual(gaVariantSet.id, variantSet.getId())
        self.assertEqual(gaVariantSet.dataset_id, dataset.getId())
        self.assertEqual(gaVariantSet.name, variantSet.getLocalId())
        self.assertItemsEqual(gaVariantSet.metadata, variantSet.getMetadata())

    def verifyCallSetsEqual(self, gaCallSet, callSet):
        variantSet = callSet.getParentContainer()
        self.assertEqual(gaCallSet.id, callSet.getId())
        self.assertEqual(gaCallSet.name, callSet.getLocalId())
        self.assertEqual(gaCallSet.variant_set_ids, [variantSet.getId()])
        for key, value in gaCallSet.info.items():
            self.assertEqual(value[0], callSet.getInfo()[key])

    def verifyReadGroupSetsEqual(self, gaReadGroupSet, readGroupSet):
        dataset = readGroupSet.getParentContainer()
        self.assertEqual(gaReadGroupSet.id, readGroupSet.getId())
        self.assertEqual(gaReadGroupSet.dataset_id, dataset.getId())
        self.assertEqual(gaReadGroupSet.name, readGroupSet.getLocalId())
        self.assertEqual(
            len(gaReadGroupSet.read_groups), len(readGroupSet.getReadGroups()))
        for gaReadGroup, readGroup in zip(
                gaReadGroupSet.read_groups, readGroupSet.getReadGroups()):
            self.verifyReadGroupsEqual(gaReadGroup, readGroup)

    def verifyReadGroupsEqual(self, gaReadGroup, readGroup):
        self.assertEqual(gaReadGroup.id, readGroup.getId())

    def verifyDatasetsEqual(self, gaDataset, dataset):
        self.assertEqual(gaDataset.id, dataset.getId())
        self.assertEqual(gaDataset.name, dataset.getLocalId())
        self.assertEqual(gaDataset.description, dataset.getDescription())

    def verifyReferenceSetsEqual(self, gaReferenceSet, referenceSet):
        self.assertEqual(gaReferenceSet.id, referenceSet.getId())
        self.assertEqual(
            gaReferenceSet.md5checksum, referenceSet.getMd5Checksum())
        self.assertEqual(
            gaReferenceSet.ncbi_taxon_id, referenceSet.getNcbiTaxonId())
        self.assertEqual(
            gaReferenceSet.assembly_id, referenceSet.getAssemblyId())
        self.assertEqual(
            gaReferenceSet.source_uri, referenceSet.getSourceUri())
        self.assertEqual(
            gaReferenceSet.source_accessions,
            referenceSet.getSourceAccessions())
        self.assertEqual(
            gaReferenceSet.is_derived, referenceSet.getIsDerived())
        self.assertEqual(
            gaReferenceSet.name, referenceSet.getLocalId())

    def verifyFeatureSetsEqual(self, gaFeatureSet, featureSet):
        dataset = featureSet.getParentContainer()
        self.assertEqual(gaFeatureSet.id, featureSet.getId())
        self.assertEqual(gaFeatureSet.dataset_id, dataset.getId())
        self.assertEqual(gaFeatureSet.name, featureSet.getLocalId())

    def verifyFeaturesEquivalent(self, f1, f2):
        # at least modulo featureId. They can obviously have different
        # start/end/etc parameters if randomly generated from search vs get.
        self.assertEqual(f1.id, f2.id)
        self.assertEqual(f1.parent_id, f2.parent_id)
        self.assertEqual(f1.feature_set_id, f2.feature_set_id)

    def verifyReferencesEqual(self, gaReference, reference):
        self.assertEqual(gaReference.id, reference.getId())
        self.assertEqual(gaReference.name, reference.getName())
        self.assertEqual(gaReference.length, reference.getLength())
        self.assertEqual(gaReference.md5checksum, reference.getMd5Checksum())
        self.assertEqual(gaReference.ncbi_taxon_id, reference.getNcbiTaxonId())
        self.assertEqual(gaReference.source_uri, reference.getSourceUri())
        self.assertEqual(
            gaReference.source_accessions, reference.getSourceAccessions())
        self.assertEqual(gaReference.is_derived, reference.getIsDerived())
        self.assertEqual(
            gaReference.source_divergence, reference.getSourceDivergence())

    def verifySearchMethod(
            self, request, path, responseClass, objects, objectVerifier):
        """
        Verifies that the specified search request operates correctly
        and returns all the speficied objects. The specified verifier
        function checks that all the returned objects are equivalent
        to their datamodel counterparts.
        """
        request.page_size = len(objects)
        self.assertGreater(request.page_size, 0)
        responseData = self.sendSearchRequest(path, request, responseClass)
        self.assertEqual(responseData.next_page_token, "")
        responseList = getattr(
            responseData, protocol.getValueListName(responseClass))
        self.assertEqual(len(objects), len(responseList))
        for gaObject, datamodelObject in zip(responseList, objects):
            objectVerifier(gaObject, datamodelObject)

    def verifySearchResultsEmpty(self, request, path, responseClass):
        """
        Verifies that we get a successful response with an empty list of
        results.
        """
        responseData = self.sendSearchRequest(path, request, responseClass)
        self.assertEqual("", responseData.next_page_token)
        responseList = getattr(
            responseData, protocol.getValueListName(responseClass))
        self.assertEqual(0, len(responseList))

    def assertObjectNotFound(self, response):
        """
        Checks that the specified response contains a search failure.
        """
        self.assertEqual(404, response.status_code)
        error = protocol.fromJson(response.data, protocol.GAException)
        self.assertTrue(protocol.validate(protocol.toJson(error), type(error)))
        self.assertGreater(error.error_code, 0)
        self.assertGreater(len(error.message), 0)

    def assertObjectNotSupported(self, response):
        """
        Checks that the specified response returns a not supported 501 status
        """
        self.assertEqual(501, response.status_code)
        error = protocol.fromJson(response.data, protocol.GAException)
        self.assertTrue(protocol.validate(protocol.toJson(error), type(error)))
        self.assertGreater(error.error_code, 0)
        self.assertGreater(len(error.message), 0)

    def verifySearchMethodFails(self, request, path):
        """
        Verify that the specified search request fails with a 404.
        """
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        self.assertObjectNotFound(response)

    def verifySearchMethodNotSupported(self, request, path):
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        self.assertObjectNotSupported(response)

    def verifyGetMethodFails(self, path, id_):
        """
        Verifies the specified GET request failes with a 404.
        """
        response = self.sendObjectGetRequest(path, id_)
        self.assertObjectNotFound(response)

    def testGetDataset(self):
        path = "/datasets"
        for dataset in self.dataRepo.getDatasets():
            responseObject = self.sendGetObject(
                path, dataset.getId(), protocol.Dataset)
            self.verifyDatasetsEqual(responseObject, dataset)
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testDatasetsSearch(self):
        request = protocol.SearchDatasetsRequest()
        datasets = self.dataRepo.getDatasets()
        path = '/datasets/search'
        self.verifySearchMethod(
            request, path, protocol.SearchDatasetsResponse, datasets,
            self.verifyDatasetsEqual)

    def testVariantSetsSearch(self):
        path = '/variantsets/search'
        for dataset in self.dataRepo.getDatasets():
            variantSets = dataset.getVariantSets()
            request = protocol.SearchVariantSetsRequest()
            request.dataset_id = dataset.getId()
            self.verifySearchMethod(
                request, path, protocol.SearchVariantSetsResponse, variantSets,
                self.verifyVariantSetsEqual)
        for badId in self.getBadIds():
            request = protocol.SearchVariantSetsRequest()
            request.dataset_id = badId
            self.verifySearchMethodFails(request, path)

    def testCallSetsSearch(self):
        path = '/callsets/search'
        for dataset in self.dataRepo.getDatasets():
            for variantSet in dataset.getVariantSets():
                callSets = variantSet.getCallSets()
                self.assertGreater(len(callSets), 0)
                request = protocol.SearchCallSetsRequest()
                request.variant_set_id = variantSet.getId()
                self.verifySearchMethod(
                    request, path, protocol.SearchCallSetsResponse, callSets,
                    self.verifyCallSetsEqual)
                # Check if we can search for the callset with a good name.
                for callSet in callSets:
                    request = protocol.SearchCallSetsRequest()
                    request.variant_set_id = variantSet.getId()
                    request.name = callSet.getLocalId()
                    self.verifySearchMethod(
                        request, path, protocol.SearchCallSetsResponse,
                        [callSet], self.verifyCallSetsEqual)
                # Check if we can search for the callset with a bad name.
                for badId in self.getBadIds():
                    request = protocol.SearchCallSetsRequest()
                    request.variant_set_id = variantSet.getId()
                    request.name = badId
                    self.verifySearchResultsEmpty(
                        request, path, protocol.SearchCallSetsResponse)
        # Check for searches within missing variantSets.
        for badId in self.getBadIds():
            request = protocol.SearchCallSetsRequest()
            request.variant_set_id = badId
            self.verifySearchMethodFails(request, path)

    def testReadGroupSetsSearch(self):
        path = '/readgroupsets/search'
        for dataset in self.dataRepo.getDatasets():
            readGroupSets = dataset.getReadGroupSets()
            request = protocol.SearchReadGroupSetsRequest()
            request.dataset_id = dataset.getId()
            self.verifySearchMethod(
                request, path, protocol.SearchReadGroupSetsResponse,
                readGroupSets, self.verifyReadGroupSetsEqual)
            # Check if we can search for the readGroupSet with a good name.
            for readGroupSet in readGroupSets:
                request = protocol.SearchReadGroupSetsRequest()
                request.dataset_id = dataset.getId()
                request.name = readGroupSet.getLocalId()
                self.verifySearchMethod(
                    request, path, protocol.SearchReadGroupSetsResponse,
                    [readGroupSet], self.verifyReadGroupSetsEqual)
            # Check if we can search for the readGroupSet with a bad name.
            for badId in self.getBadIds():
                request = protocol.SearchReadGroupSetsRequest()
                request.dataset_id = dataset.getId()
                request.name = badId
                self.verifySearchResultsEmpty(
                    request, path, protocol.SearchReadGroupSetsResponse)
        for badId in self.getBadIds():
            request = protocol.SearchReadGroupSetsRequest()
            request.dataset_id = badId
            self.verifySearchMethodFails(request, path)

    def testReferenceSetsSearch(self):
        request = protocol.SearchReferenceSetsRequest()
        referenceSets = self.dataRepo.getReferenceSets()
        path = '/referencesets/search'
        self.verifySearchMethod(
            request, path, protocol.SearchReferenceSetsResponse, referenceSets,
            self.verifyReferenceSetsEqual)

    def testReferencesSearch(self):
        path = '/references/search'
        for referenceSet in self.dataRepo.getReferenceSets():
            references = referenceSet.getReferences()
            request = protocol.SearchReferencesRequest()
            request.reference_set_id = referenceSet.getId()
            self.verifySearchMethod(
                request, path, protocol.SearchReferencesResponse, references,
                self.verifyReferencesEqual)
        for badId in self.getBadIds():
            request = protocol.SearchReferencesRequest()
            request.reference_set_id = badId
            self.verifySearchMethodFails(request, path)

    def verifyReferenceSearchFilters(
            self, objectList, hasAssemblyId, path, requestFactory,
            responseClass, objectVerifier):
        """
        Verifies the filtering functionality for the specified list of
        reference-like objects.
        """
        self.assertGreater(len(objectList), 2)
        for obj in objectList[1:]:
            request = requestFactory()
            # First, check the simple cases; 1 filter set, others null.
            request.md5checksum = obj.getMd5Checksum()
            self.verifySearchMethod(
                request, path, responseClass, [obj], objectVerifier)
            request.md5checksum = ""
            request.accession = obj.getSourceAccessions()[0]
            self.verifySearchMethod(
                request, path, responseClass, [obj], objectVerifier)
            request.accession = ""
            if hasAssemblyId:
                request.assembly_id = obj.getAssemblyId()
                self.verifySearchMethod(
                    request, path, responseClass, [obj], objectVerifier)
                request.assembly_id = ""
            # Now check one good value and some bad values.
            request.md5checksum = obj.getMd5Checksum()
            badAccessions = [
                "no such accession", objectList[0].getSourceAccessions()[0]]
            for accession in badAccessions:
                request.accession = accession
                self.verifySearchResultsEmpty(request, path, responseClass)
            request.accession = ""
            if hasAssemblyId:
                badAssemblyIds = [
                    "no such asssembly", objectList[0].getAssemblyId()]
                for assemblyId in badAssemblyIds:
                    request.assembly_id = assemblyId
                    self.verifySearchResultsEmpty(request, path, responseClass)
                request.assembly_id = ""

    def testReferencesSearchFilters(self):
        path = '/references/search'
        for referenceSet in self.dataRepo.getReferenceSets():

            def requestFactory():
                request = protocol.SearchReferencesRequest()
                request.reference_set_id = referenceSet.getId()
                return request
            self.verifyReferenceSearchFilters(
                referenceSet.getReferences(), False, path, requestFactory,
                protocol.SearchReferencesResponse, self.verifyReferencesEqual)

    def testReferenceSetsSearchFilters(self):
        path = '/referencesets/search'

        def requestFactory():
            return protocol.SearchReferenceSetsRequest()
        self.verifyReferenceSearchFilters(
            self.dataRepo.getReferenceSets(), True, path, requestFactory,
            protocol.SearchReferenceSetsResponse,
            self.verifyReferenceSetsEqual)

    def testGetVariantSet(self):
        path = "/variantsets"
        for dataset in self.dataRepo.getDatasets():
            for variantSet in dataset.getVariantSets():
                responseObject = self.sendGetObject(
                    path, variantSet.getId(), protocol.VariantSet)
                self.verifyVariantSetsEqual(responseObject, variantSet)
            for badId in self.getBadIds():
                variantSet = variants.AbstractVariantSet(dataset, badId)
                self.verifyGetMethodFails(path, variantSet.getId())
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testGetVariantAnnotationSet(self):
        path = "/variantannotationsets"
        for dataset in self.dataRepo.getDatasets():
            for variantSet in dataset.getVariantSets():
                for vas in variantSet.getVariantAnnotationSets():
                    responseObject = self.sendGetObject(
                        path, vas.getId(), protocol.VariantAnnotationSet)
                    self.assertEqual(
                        vas.getId(), responseObject.id,
                        "The requested ID should match the returned")
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testGetVariant(self):
        # get a variant from the search method
        referenceName = '1'
        start = 0
        request = protocol.SearchVariantsRequest()
        request.variant_set_id = self.variantSet.getId()
        request.reference_name = referenceName
        request.start = start
        request.end = 2**16
        path = '/variants/search'
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchVariantsResponse)
        variants = responseData.variants[:10]

        # get 'the same' variant using the get method
        for variant in variants:
            path = '/variants'
            responseObject = self.sendGetObject(
                path, variant.id, protocol.Variant)
            self.assertEqual(responseObject, variant)

    def testGetReferenceSet(self):
        path = "/referencesets"
        for referenceSet in self.dataRepo.getReferenceSets():
            responseObject = self.sendGetObject(
                path, referenceSet.getId(), protocol.ReferenceSet)
            self.verifyReferenceSetsEqual(responseObject, referenceSet)
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testGetReference(self):
        path = "/references"
        for referenceSet in self.dataRepo.getReferenceSets():
            for reference in referenceSet.getReferences():
                responseObject = self.sendGetObject(
                    path, reference.getId(), protocol.Reference)
                self.verifyReferencesEqual(responseObject, reference)
            for badId in self.getBadIds():
                referenceSet = references.AbstractReferenceSet(badId)
                self.verifyGetMethodFails(path, referenceSet.getId())
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testGetCallSet(self):
        path = "/callsets"
        for dataset in self.dataRepo.getDatasets():
            for variantSet in dataset.getVariantSets():
                for callSet in variantSet.getCallSets():
                    responseObject = self.sendGetObject(
                        path, callSet.getId(), protocol.CallSet)
                    self.verifyCallSetsEqual(responseObject, callSet)
                for badId in self.getBadIds():
                    callSet = variants.CallSet(variantSet, badId)
                    self.verifyGetMethodFails(path, callSet.getId())
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testGetReadGroup(self):
        path = "/readgroups"
        for dataset in self.dataRepo.getDatasets():
            for readGroupSet in dataset.getReadGroupSets():
                for readGroup in readGroupSet.getReadGroups():
                    responseObject = self.sendGetObject(
                        path, readGroup.getId(), protocol.ReadGroup)
                    self.verifyReadGroupsEqual(responseObject, readGroup)
                for badId in self.getBadIds():
                    readGroup = reads.AbstractReadGroup(readGroupSet, badId)
                    self.verifyGetMethodFails(path, readGroup.getId())
            for badId in self.getBadIds():
                readGroupSet = reads.AbstractReadGroupSet(dataset, badId)
                self.verifyGetMethodFails(path, readGroupSet.getId())
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testVariantsSearch(self):
        referenceName = '1'

        request = protocol.SearchVariantsRequest()
        request.reference_name = referenceName
        request.start = 0
        request.end = 0
        request.variant_set_id = self.variantSet.getId()

        # Request windows is too small, no results
        path = '/variants/search'
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchVariantsResponse)
        self.assertEqual("", responseData.next_page_token)
        self.assertEqual(0, len(responseData.variants))

        # Larger request window, expect results
        request.end = 2 ** 16
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchVariantsResponse)
        self.assertTrue(protocol.validate(
            protocol.toJson(responseData), protocol.SearchVariantsResponse))
        self.assertGreater(len(responseData.variants), 0)

        # Verify all results are in the correct range, set and reference
        for variant in responseData.variants:
            self.assertGreaterEqual(variant.start, 0)
            self.assertLessEqual(variant.end, 2 ** 16)
            self.assertEqual(variant.variant_set_id, self.variantSet.getId())
            self.assertEqual(variant.reference_name, referenceName)

        # TODO: Add more useful test scenarios, including some covering
        # pagination behavior.

    def testVariantAnnotationSetsSearch(self):
        self.assertIsNotNone(self.variantAnnotationSet)

        request = protocol.SearchVariantAnnotationSetsRequest()

        request.variant_set_id = "b4d=="
        path = '/variantannotationsets/search'
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        responseData = protocol.fromJson(response.data, protocol.GAException)
        self.assertTrue(protocol.validate(protocol.toJson(responseData),
                                          protocol.GAException))
        self.assertEqual(responseData.error_code, 758389611)
        self.assertEqual(responseData.message,
                         'No object of this type exists with id \'b4d==\'')

        request.variant_set_id = self.variantSet.getId()
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        responseData = protocol.fromJson(response.data, protocol.
                                         SearchVariantAnnotationSetsResponse)
        self.assertTrue(protocol.validate(
            protocol.toJson(responseData),
            protocol.SearchVariantAnnotationSetsResponse))
        self.assertGreater(len(responseData.variant_annotation_sets), 0,
                           "Expect some results for a known good ID")
        # TODO check the instance variables; we should be able to match
        # the values from the protocol object we get back with the values
        # in the original variantAnnotationSet.

    def testVariantAnnotationsSearch(self):
        self.assertIsNotNone(self.variantAnnotationSet)

        request = protocol.SearchVariantAnnotationsRequest()
        # TODO split these into separate tests, and factor out the duplicated
        # code.

        path = '/variantannotations/search'
        request.start = 0
        request.end = 1000000
        request.page_size = 1
        request.reference_name = "1"
        request.variant_annotation_set_id = self.variantAnnotationSet.getId()
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        responseData = protocol.fromJson(response.data, protocol.
                                         SearchVariantAnnotationsResponse)
        self.assertGreater(len(responseData.variant_annotations), 0)
        self.assertIsNotNone(
            responseData.next_page_token,
            "Expected more than one page of results")

        request = protocol.SearchVariantAnnotationsRequest()
        request.variant_annotation_set_id = self.variantAnnotationSet.getId()
        request.start = 0
        request.end = 10
        request.reference_name = "1"

        request.effects.add().id = "ThisIsNotAnEffect"

        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        responseData = protocol.fromJson(response.data, protocol.
                                         SearchVariantAnnotationsResponse)
        self.assertEquals(
            len(responseData.variant_annotations), 0,
            "There should be no results for a nonsense effect")

        request = protocol.SearchVariantAnnotationsRequest()
        request.variant_annotation_set_id = self.variantAnnotationSet.getId()
        request.start = 0
        request.end = 10
        request.reference_name = "1"
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        responseData = protocol.fromJson(response.data, protocol.
                                         SearchVariantAnnotationsResponse)
        self.assertGreater(len(responseData.variant_annotations), 0)
        for ann in responseData.variant_annotations:
            self.assertGreater(
                len(ann.transcript_effects), 0,
                ("When no effects are requested ensure "
                    "some transcript effects are still present"))

        request = protocol.SearchVariantAnnotationsRequest()
        request.variant_annotation_set_id = self.variantAnnotationSet.getId()
        request.start = 0
        request.end = 5
        request.reference_name = "1"
        request.effects.add().id = "SO:0001627"
        request.effects.add().id = "B4DID"
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        responseData = protocol.fromJson(response.data, protocol.
                                         SearchVariantAnnotationsResponse)
        responseLength = len(responseData.variant_annotations)
        self.assertGreater(
            responseLength, 0,
            "There should be some results for a known effect")
        for ann in responseData.variant_annotations:
            effectPresent = False
            for effect in ann.transcript_effects:
                for featureType in effect.effects:
                    if featureType.id in map(
                            lambda e: e.id, request.effects):
                        effectPresent = True
            self.assertEquals(
                True, effectPresent,
                "The ontology term should appear at least once")

        request = protocol.SearchVariantAnnotationsRequest()
        request.variant_annotation_set_id = self.variantAnnotationSet.getId()
        request.start = 0
        request.end = 5
        request.reference_name = "1"
        request.effects.add().id = "B4DID"
        request.effects.add().id = "SO:0001627"
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        responseData = protocol.fromJson(response.data, protocol.
                                         SearchVariantAnnotationsResponse)
        self.assertEqual(
            len(responseData.variant_annotations),
            responseLength,
            "Order shall not affect results")
        for ann in responseData.variant_annotations:
            effectPresent = False
            for effect in ann.transcript_effects:
                for featureType in effect.effects:
                    if featureType.id in map(
                            lambda e: e.id, request.effects):
                        effectPresent = True
            self.assertEquals(
                True,
                effectPresent,
                "The ontology term should appear at least once")

        request = protocol.SearchVariantAnnotationsRequest()
        request.variant_annotation_set_id = self.variantAnnotationSet.getId()
        request.start = 0
        request.end = 5
        request.reference_name = "1"
        request.effects.add().id = "SO:0001627"
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        responseData = protocol.fromJson(response.data, protocol.
                                         SearchVariantAnnotationsResponse)
        self.assertGreater(len(responseData.variant_annotations), 0,
                           "There should be some results for a good effect ID")
        for ann in responseData.variant_annotations:
            effectPresent = False
            txIds = map(lambda t: t.id, ann.transcript_effects)
            self.assertEqual(len(txIds), len(set(txIds)),
                             "Transcript effects should be unique")
            for effect in ann.transcript_effects:
                for featureType in effect.effects:
                    if featureType.id in map(
                            lambda e: e.id, request.effects):
                        effectPresent = True
            self.assertEquals(True, effectPresent,
                              "The ontology term should appear at least once")

        request = protocol.SearchVariantAnnotationsRequest()
        request.variant_annotation_set_id = self.variantAnnotationSet.getId()
        request.start = 0
        request.end = 10
        request.reference_name = "1"
        request.effects.add().id = "SO:0001627"
        request.effects.add().id = "SO:0001791"
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        responseData = protocol.fromJson(response.data, protocol.
                                         SearchVariantAnnotationsResponse)
        self.assertGreater(len(responseData.variant_annotations), 0)

    def testGetFeatureSet(self):
        path = "/featuresets"
        for dataset in self.dataRepo.getDatasets():
            for featureSet in dataset.getFeatureSets():
                responseObject = self.sendGetObject(
                    path, featureSet.getId(), protocol.FeatureSet)
                self.verifyFeatureSetsEqual(responseObject, featureSet)
            for badId in self.getBadIds():
                featureSet = sequence_annotations.AbstractFeatureSet(
                    dataset, badId)
                self.verifyGetMethodFails(path, featureSet.getId())
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testFeatureSetsSearch(self):
        path = '/featuresets/search'
        for dataset in self.dataRepo.getDatasets():
            featureSets = dataset.getFeatureSets()
            request = protocol.SearchFeatureSetsRequest()
            request.dataset_id = dataset.getId()
            self.verifySearchMethod(
                request, path, protocol.SearchFeatureSetsResponse, featureSets,
                self.verifyFeatureSetsEqual)
        for badId in self.getBadIds():
            request = protocol.SearchFeatureSetsRequest()
            request.dataset_id = badId
            self.verifySearchMethodFails(request, path)

    def testGetFeature(self):
        dataset = self.dataRepo.getDatasets()[0]
        featureSet = dataset.getFeatureSets()[0]
        request = protocol.SearchFeaturesRequest()
        request.feature_set_id = featureSet.getId()
        request.reference_name = "chr1"
        request.start = 0
        request.end = 2**16
        path = '/features/search'
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchFeaturesResponse)
        features = responseData.features[:10]

        # get 'the same' feature using the get method
        for feature in features:
            path = '/features'
            responseObject = self.sendGetObject(
                path, feature.id, protocol.Feature)
            self.verifyFeaturesEquivalent(responseObject, feature)

    def testFeaturesSearch(self):
        dataset = self.dataRepo.getDatasets()[0]
        featureSet = dataset.getFeatureSets()[0]
        referenceName = 'chr1'

        request = protocol.SearchFeaturesRequest()
        request.reference_name = referenceName
        request.feature_set_id = featureSet.getId()

        # Request window is outside of simulated dataset bounds, no results
        request.start = 0
        request.end = 1
        request.parent_id = ''
        path = '/features/search'
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchFeaturesResponse)
        self.assertEqual('', responseData.next_page_token)
        self.assertEqual(0, len(responseData.features))

        # Larger request window, expect results
        request.start = 0
        request.end = 2 ** 16
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchFeaturesResponse)

        self.assertGreater(len(responseData.features), 0)

        # Verify all results are in the correct range, set and reference
        for feature in responseData.features:
            self.assertGreaterEqual(feature.start, 0)
            self.assertLessEqual(feature.end, 2 ** 16)
            self.assertEqual(feature.feature_set_id, featureSet.getId())
            self.assertEqual(feature.reference_name, referenceName)

    def testListReferenceBases(self):
        for referenceSet in self.dataRepo.getReferenceSets():
            for reference in referenceSet.getReferences():
                length = reference.getLength()
                sequence = reference.getBases(0, length)
                # fetch the bases
                args = protocol.ListReferenceBasesRequest()
                args.reference_id = reference.getId()
                response = self.sendListReferenceBasesRequest(args)
                self.assertEqual(response.sequence, sequence)
                # Try some simple slices.
                ranges = [(0, length), (0, 1), (length - 1, length)]
                for start, end in ranges:
                    args = protocol.ListReferenceBasesRequest()
                    args.start, args.end = start, end
                    args.reference_id = reference.getId()
                    response = self.sendListReferenceBasesRequest(args)
                    self.assertEqual(response.sequence, sequence[start:end])
                    self.assertEqual("", response.next_page_token)
                    self.assertEqual(response.offset, start)

    def testListReferenceBasesErrors(self):
        referenceSet = self.dataRepo.getReferenceSets()[0]
        args = protocol.ListReferenceBasesRequest()
        for badId in self.getBadIds():
            args.reference_id = badId
            path = '/listreferencebases'
            response = self.sendJsonPostRequest(path, protocol.toJson(args))
            self.assertEqual(response.status_code, 404)
            reference = references.AbstractReference(referenceSet, badId)
            args.reference_id = reference.getId()
            path = '/listreferencebases'
            response = self.sendJsonPostRequest(path, protocol.toJson(args))
            self.assertEqual(response.status_code, 404)
        path = '/listreferencebases'
        length = self.reference.getLength()
        badRanges = [(-1, 0), (-1, -1), (length, 0), (0, length + 1)]
        for start, end in badRanges:
            args = protocol.ListReferenceBasesRequest()
            args.start, args.end = start, end
            args.reference_id = self.reference.getId()
            response = self.sendJsonPostRequest(path, protocol.toJson(args))
            self.assertEqual(response.status_code, 416)

    def testListReferenceBasesPaging(self):
        id_ = self.reference.getId()
        length = self.reference.getLength()
        completeSequence = self.reference.getBases(0, length)
        for start, end in [(0, length), (5, 10), (length // 2, length)]:
            sequence = completeSequence[start: end]
            for pageSize in [1, 2, length - 1]:
                self.backend.setMaxResponseLength(pageSize)
                args = protocol.ListReferenceBasesRequest()
                args.start, args.end, args.reference_id = start, end, id_
                response = self.sendListReferenceBasesRequest(args)
                self.assertEqual(response.sequence, sequence[:pageSize])
                self.assertEqual(response.offset, start)
                sequenceFragments = [response.sequence]
                while response.next_page_token:
                    args = protocol.ListReferenceBasesRequest()
                    args.page_token = response.next_page_token
                    args.start, args.end, args.reference_id = start, end, id_
                    response = self.sendListReferenceBasesRequest(args)
                    self.assertGreater(len(response.sequence), 0)
                    sequenceFragments.append(response.sequence)
                    offset = response.offset
                    self.assertEqual(
                        response.sequence,
                        completeSequence[
                            offset: offset + len(response.sequence)])

                self.assertEqual("".join(sequenceFragments), sequence)

    def testReads(self):
        path = '/reads/search'
        for dataset in self.dataRepo.getDatasets():
            for readGroupSet in dataset.getReadGroupSets():
                referenceSet = readGroupSet.getReferenceSet()
                for reference in referenceSet.getReferences():
                    for readGroup in readGroupSet.getReadGroups():
                        # search reads
                        request = protocol.SearchReadsRequest()
                        request.read_group_ids.append(readGroup.getId())
                        request.reference_id = reference.getId()
                        responseData = self.sendSearchRequest(
                            path, request, protocol.SearchReadsResponse)
                        alignments = responseData.alignments
                        self.assertGreater(len(alignments), 0)
                        for alignment in alignments:
                            # TODO more tests here: this is very weak.
                            self.assertEqual(
                                alignment.read_group_id, readGroup.getId())

    def testUnsupportedReadOperations(self):
        path = '/reads/search'

        # unmapped Reads
        request = protocol.SearchReadsRequest()
        request.read_group_ids.extend([self.readGroup.getId()])
        request.reference_id = ""
        self.verifySearchMethodNotSupported(request, path)

        # multiple ReadGroupSets set mismatch
        request.read_group_ids.append(self.readGroup.getId())
        request.read_group_ids.append("42")
        request.reference_id = self.reference.getId()
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        self.assertEqual(400, response.status_code)

    def testReadsMultipleReadGroupSets(self):
        path = '/reads/search'
        readGroupIds = [
            readGroup.getId() for readGroup in
            self.readGroupSet.getReadGroups()]
        referenceId = self.reference.getId()

        request = protocol.SearchReadsRequest()
        request.read_group_ids.extend(readGroupIds)
        request.reference_id = referenceId
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchReadsResponse)

        readGroupAlignments = []
        for readGroup in self.readGroupSet.getReadGroups():
            readGroupAlignments.extend(list(
                readGroup.getReadAlignments(referenceId, None, None)))

        alignments = responseData.alignments
        self.assertEqual(len(alignments), len(readGroupAlignments))
        for alignment, rgAlignment in zip(alignments, readGroupAlignments):
            self.assertEqual(alignment.id, rgAlignment.id)
            self.assertEqual(
                alignment.read_group_id,
                rgAlignment.read_group_id)

    def testBioSamplesFromReadGroupSets(self):
        path = 'readgroupsets/search'
        dataset = self.dataRepo.getDatasets()[0]
        # get all the read group sets
        request = protocol.SearchReadGroupSetsRequest()
        request.dataset_id = dataset.getId()
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchReadGroupSetsResponse)
        # go through each read group
        bioSamplesRgs = []
        ran = False
        for rgs in responseData.read_group_sets:
            for rg in rgs.read_groups:
                ran = True
                # request biosample record
                if rg.bio_sample_id:
                    bioSample = self.sendGetObject(
                        'biosamples',
                        rg.bio_sample_id,
                        protocol.BioSample)
                    bioSamplesRgs.append((bioSample.id, rgs.id))
                    self.assertNotEqual(
                        None, bioSample,
                        "A BioSample should exist for reach read")
        self.assertTrue(ran)
        # search reads by biosample
        ran = False
        for bsId, rgsId in bioSamplesRgs:
            request = protocol.SearchReadGroupSetsRequest()
            request.dataset_id = dataset.getId()
            request.bio_sample_id = bsId
            request.name = "A BAD NAME"
            request.page_size = 1
            responseData = self.sendSearchRequest(
                path, request, protocol.SearchReadGroupSetsResponse)
            self.assertEquals(
                len(responseData.read_group_sets), 0,
                "A good biosample ID and bad name should return 0")
            request = protocol.SearchReadGroupSetsRequest()
            request.dataset_id = dataset.getId()
            request.bio_sample_id = bsId
            responseData = self.sendSearchRequest(
                path, request, protocol.SearchReadGroupSetsResponse)
            for rgs in responseData.read_group_sets:
                for rg in rgs.read_groups:
                    ran = True
                    self.assertEqual(
                        rg.bio_sample_id, bsId,
                        "Only read groups matching the BioSample ID")
        self.assertTrue(ran)

    def testBioSamplesFromCallSets(self):
        path = 'callsets/search'
        dataset = self.dataRepo.getDatasets()[0]
        variantSet = dataset.getVariantSets()[0]
        callSet = variantSet.getCallSets()[0]
        request = protocol.SearchCallSetsRequest()
        request.variant_set_id = variantSet.getId()
        request.bio_sample_id = "A BAD ID"
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchCallSetsResponse)
        self.assertEqual(len(responseData.call_sets), 0)

        request = protocol.SearchCallSetsRequest()
        request.variant_set_id = variantSet.getId()
        request.bio_sample_id = callSet.toProtocolElement().bio_sample_id
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchCallSetsResponse)
        self.assertGreater(len(responseData.call_sets), 0)
        for cs in responseData.call_sets:
            self.assertEqual(cs.bio_sample_id, request.bio_sample_id)

        request = protocol.SearchCallSetsRequest()
        request.variant_set_id = variantSet.getId()
        request.bio_sample_id = callSet.toProtocolElement().bio_sample_id
        request.name = "A BAD NAME"
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchCallSetsResponse)
        self.assertEqual(
            len(responseData.call_sets), 0,
            "None should be returned")

    def testBioSamplesSearch(self):
        path = 'biosamples/search'
        dataset = self.dataRepo.getDatasets()[0]
        request = protocol.SearchBioSamplesRequest()
        request.name = "BAD NAME"
        request.dataset_id = dataset.getId()
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchBioSamplesResponse)
        self.assertEqual(
            len(responseData.biosamples), 0,
            "A bad name should return none")
        request = protocol.SearchBioSamplesRequest()
        request.name = "simCallSet_0"
        request.dataset_id = dataset.getId()
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchBioSamplesResponse)
        # Currently always returns a singleton
        self.assertGreater(
            len(responseData.biosamples), 0,
            "A good name should return some")

        request = protocol.SearchBioSamplesRequest()
        request.individual_id = "BAD ID"
        request.dataset_id = dataset.getId()
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchBioSamplesResponse)
        self.assertEqual(
            len(responseData.biosamples), 0,
            "A bad individual ID should return none")

        request = protocol.SearchIndividualsRequest()
        request.dataset_id = dataset.getId()
        responseData = self.sendSearchRequest(
            "individuals/search", request, protocol.SearchIndividualsResponse)
        self.assertGreater(
            len(responseData.individuals), 0,
            "Some individuals should be returned")

        individualId = responseData.individuals[0].id

        request = protocol.SearchBioSamplesRequest()
        request.individual_id = individualId
        request.dataset_id = dataset.getId()
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchBioSamplesResponse)
        self.assertGreater(
            len(responseData.biosamples), 0,
            "A good individual ID should return some")

        request = protocol.SearchBioSamplesRequest()
        request.individual_id = individualId
        request.page_size = 1
        request.dataset_id = dataset.getId()
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchBioSamplesResponse)
        self.assertIsNotNone(
            responseData.next_page_token,
            "More than one page should be returned")
        request.page_token = responseData.next_page_token
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchBioSamplesResponse)
        self.assertEqual(
            responseData.biosamples[0].individual_id,
            individualId,
            "Results on the second page should match")

    def testSearchIndividuals(self):
        path = 'individuals/search'
        dataset = self.dataRepo.getDatasets()[0]
        request = protocol.SearchIndividualsRequest()
        request.name = "BAD NAME"
        request.dataset_id = dataset.getId()
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchIndividualsResponse)
        self.assertEqual(
            len(responseData.individuals), 0,
            "A bad individual name should return none")
        request = protocol.SearchIndividualsRequest()
        request.name = "simCallSet_0"
        request.dataset_id = dataset.getId()
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchIndividualsResponse)
        self.assertGreater(
            len(responseData.individuals), 0,
            "A good individual name should return some")

    def testGetIndividual(self):
        path = "/individuals"
        for dataset in self.dataRepo.getDatasets():
            for individual in dataset.getIndividuals():
                responseObject = self.sendGetObject(
                    path,
                    individual.getId(),
                    protocol.Individual)
                self.assertEqual(
                    responseObject.id, individual.getId())
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testGetBioSample(self):
        path = "/biosamples"
        for dataset in self.dataRepo.getDatasets():
            for bioSample in dataset.getBioSamples():
                responseObject = self.sendGetObject(
                    path,
                    bioSample.getId(),
                    protocol.BioSample)
                self.assertEqual(responseObject.id, bioSample.getId())
        for badId in self.getBadIds():
            self.verifyGetMethodFails(path, badId)

    def testSearchPhenotypeAssociationSets(self):
        path = "/phenotypeassociationsets/search"
        for dataset in self.dataRepo.getDatasets():
            repoPaSetIds = []
            for repoPaSet in dataset.getPhenotypeAssociationSets():
                repoPaSetIds.append(repoPaSet.getId())
            request = protocol.SearchPhenotypeAssociationSetsRequest()
            request.dataset_id = dataset.getId()
            responseData = self.sendSearchRequest(
                path, request,
                protocol.SearchPhenotypeAssociationSetsResponse)
            for clientPaSet in responseData.phenotype_association_sets:
                self.assertTrue(clientPaSet.id in repoPaSetIds)

    def testSearchPhenotypes(self):
        path = "/phenotypes/search"
        for repoPaSet in self.dataRepo.allPhenotypeAssociationSets():
            for repoAssoc in repoPaSet.getAssociations():
                request = protocol.SearchPhenotypesRequest()
                request.phenotype_association_set_id = repoPaSet.getId()
                request.id = repoAssoc.phenotype.id
                responseData = self.sendSearchRequest(
                    path, request,
                    protocol.SearchPhenotypesResponse)
                for clientPhenotype in responseData.phenotypes:
                    self.assertEqual(clientPhenotype, repoAssoc.phenotype)

    def testSearchGenotypePhenotypes(self):
        path = "/featurephenotypeassociations/search"
        for repoPaSet in self.dataRepo.allPhenotypeAssociationSets():
            for repoAssoc in repoPaSet.getAssociations():
                request = protocol.SearchGenotypePhenotypeRequest()
                request.phenotype_association_set_id = repoPaSet.getId()
                request.phenotype_ids.extend([repoAssoc.phenotype.id])
                responseData = self.sendSearchRequest(
                    path, request,
                    protocol.SearchGenotypePhenotypeResponse)
                for clientAssoc in responseData.associations:
                    self.assertEqual(clientAssoc, repoAssoc)

    # TODO def testSearchGenotypePhenotypes(self):

    # TODO def testGetExpressionLevel(self):

    # TODO def testSearchExpressionLevels(self):

    # TODO def testGetRnaQuantification(self):

    # TODO def testSearchRnaQuantifications(self):

    # TODO def testGetRnaQuantificationSet(self):

    # TODO def testSearchRnaQuantificationSets(self):



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ga4gh/reads.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ga4gh import common_pb2 as ga4gh_dot_common__pb2
from ga4gh import assay_metadata_pb2 as ga4gh_dot_assay__metadata__pb2
from ga4gh import metadata_pb2 as ga4gh_dot_metadata__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ga4gh/reads.proto',
  package='ga4gh',
  syntax='proto3',
  serialized_pb=_b('\n\x11ga4gh/reads.proto\x12\x05ga4gh\x1a\x12ga4gh/common.proto\x1a\x1aga4gh/assay_metadata.proto\x1a\x14ga4gh/metadata.proto\x1a\x1cgoogle/protobuf/struct.proto\"Y\n\tReadStats\x12\x1a\n\x12\x61ligned_read_count\x18\x01 \x01(\x03\x12\x1c\n\x14unaligned_read_count\x18\x02 \x01(\x03\x12\x12\n\nbase_count\x18\x03 \x01(\x03\"\xb2\x03\n\tReadGroup\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\ndataset_id\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x13\n\x0bsample_name\x18\x05 \x01(\t\x12\x15\n\rbio_sample_id\x18\x06 \x01(\t\x12%\n\nexperiment\x18\x07 \x01(\x0b\x32\x11.ga4gh.Experiment\x12\x1d\n\x15predicted_insert_size\x18\x08 \x01(\x05\x12\x0f\n\x07\x63reated\x18\t \x01(\x03\x12\x0f\n\x07updated\x18\n \x01(\x03\x12\x1f\n\x05stats\x18\x0b \x01(\x0b\x32\x10.ga4gh.ReadStats\x12 \n\x08programs\x18\x0c \x03(\x0b\x32\x0e.ga4gh.Program\x12\x18\n\x10reference_set_id\x18\r \x01(\t\x12(\n\x04info\x18\x0e \x03(\x0b\x32\x1a.ga4gh.ReadGroup.InfoEntry\x1aG\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.ListValue:\x02\x38\x01\"\x84\x01\n\x0cReadGroupSet\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\ndataset_id\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1f\n\x05stats\x18\x04 \x01(\x0b\x32\x10.ga4gh.ReadStats\x12%\n\x0bread_groups\x18\x05 \x03(\x0b\x32\x10.ga4gh.ReadGroup\"n\n\x0fLinearAlignment\x12!\n\x08position\x18\x01 \x01(\x0b\x32\x0f.ga4gh.Position\x12\x17\n\x0fmapping_quality\x18\x02 \x01(\x05\x12\x1f\n\x05\x63igar\x18\x03 \x03(\x0b\x32\x10.ga4gh.CigarUnit\"\xab\x04\n\rReadAlignment\x12\n\n\x02id\x18\x01 \x01(\t\x12\x15\n\rread_group_id\x18\x02 \x01(\t\x12\x15\n\rfragment_name\x18\x03 \x01(\t\x12\x1a\n\x12improper_placement\x18\x04 \x01(\x08\x12\x1a\n\x12\x64uplicate_fragment\x18\x05 \x01(\x08\x12\x14\n\x0cnumber_reads\x18\x06 \x01(\x05\x12\x17\n\x0f\x66ragment_length\x18\x07 \x01(\x05\x12\x13\n\x0bread_number\x18\x08 \x01(\x05\x12$\n\x1c\x66\x61iled_vendor_quality_checks\x18\t \x01(\x08\x12)\n\talignment\x18\n \x01(\x0b\x32\x16.ga4gh.LinearAlignment\x12\x1b\n\x13secondary_alignment\x18\x0b \x01(\x08\x12\x1f\n\x17supplementary_alignment\x18\x0c \x01(\x08\x12\x18\n\x10\x61ligned_sequence\x18\r \x01(\t\x12\x17\n\x0f\x61ligned_quality\x18\x0e \x03(\x05\x12+\n\x12next_mate_position\x18\x0f \x01(\x0b\x32\x0f.ga4gh.Position\x12,\n\x04info\x18\x10 \x03(\x0b\x32\x1e.ga4gh.ReadAlignment.InfoEntry\x1aG\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.ListValue:\x02\x38\x01\"\xa2\x02\n\tCigarUnit\x12-\n\toperation\x18\x01 \x01(\x0e\x32\x1a.ga4gh.CigarUnit.Operation\x12\x18\n\x10operation_length\x18\x02 \x01(\x03\x12\x1a\n\x12reference_sequence\x18\x03 \x01(\t\"\xaf\x01\n\tOperation\x12\x19\n\x15OPERATION_UNSPECIFIED\x10\x00\x12\x13\n\x0f\x41LIGNMENT_MATCH\x10\x01\x12\n\n\x06INSERT\x10\x02\x12\n\n\x06\x44\x45LETE\x10\x03\x12\x08\n\x04SKIP\x10\x04\x12\r\n\tCLIP_SOFT\x10\x05\x12\r\n\tCLIP_HARD\x10\x06\x12\x07\n\x03PAD\x10\x07\x12\x12\n\x0eSEQUENCE_MATCH\x10\x08\x12\x15\n\x11SEQUENCE_MISMATCH\x10\tb\x06proto3')
  ,
  dependencies=[ga4gh_dot_common__pb2.DESCRIPTOR,ga4gh_dot_assay__metadata__pb2.DESCRIPTOR,ga4gh_dot_metadata__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)



_CIGARUNIT_OPERATION = _descriptor.EnumDescriptor(
  name='Operation',
  full_name='ga4gh.CigarUnit.Operation',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OPERATION_UNSPECIFIED', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ALIGNMENT_MATCH', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INSERT', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DELETE', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SKIP', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIP_SOFT', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIP_HARD', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAD', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SEQUENCE_MATCH', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SEQUENCE_MISMATCH', index=9, number=9,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1577,
  serialized_end=1752,
)
_sym_db.RegisterEnumDescriptor(_CIGARUNIT_OPERATION)


_READSTATS = _descriptor.Descriptor(
  name='ReadStats',
  full_name='ga4gh.ReadStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='aligned_read_count', full_name='ga4gh.ReadStats.aligned_read_count', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='unaligned_read_count', full_name='ga4gh.ReadStats.unaligned_read_count', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='base_count', full_name='ga4gh.ReadStats.base_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=128,
  serialized_end=217,
)


_READGROUP_INFOENTRY = _descriptor.Descriptor(
  name='InfoEntry',
  full_name='ga4gh.ReadGroup.InfoEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ga4gh.ReadGroup.InfoEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='ga4gh.ReadGroup.InfoEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=583,
  serialized_end=654,
)

_READGROUP = _descriptor.Descriptor(
  name='ReadGroup',
  full_name='ga4gh.ReadGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='ga4gh.ReadGroup.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dataset_id', full_name='ga4gh.ReadGroup.dataset_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='ga4gh.ReadGroup.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='description', full_name='ga4gh.ReadGroup.description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sample_name', full_name='ga4gh.ReadGroup.sample_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bio_sample_id', full_name='ga4gh.ReadGroup.bio_sample_id', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='experiment', full_name='ga4gh.ReadGroup.experiment', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='predicted_insert_size', full_name='ga4gh.ReadGroup.predicted_insert_size', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='created', full_name='ga4gh.ReadGroup.created', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='updated', full_name='ga4gh.ReadGroup.updated', index=9,
      number=10, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stats', full_name='ga4gh.ReadGroup.stats', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='programs', full_name='ga4gh.ReadGroup.programs', index=11,
      number=12, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='reference_set_id', full_name='ga4gh.ReadGroup.reference_set_id', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='info', full_name='ga4gh.ReadGroup.info', index=13,
      number=14, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_READGROUP_INFOENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=220,
  serialized_end=654,
)


_READGROUPSET = _descriptor.Descriptor(
  name='ReadGroupSet',
  full_name='ga4gh.ReadGroupSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='ga4gh.ReadGroupSet.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dataset_id', full_name='ga4gh.ReadGroupSet.dataset_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='ga4gh.ReadGroupSet.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stats', full_name='ga4gh.ReadGroupSet.stats', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='read_groups', full_name='ga4gh.ReadGroupSet.read_groups', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=657,
  serialized_end=789,
)


_LINEARALIGNMENT = _descriptor.Descriptor(
  name='LinearAlignment',
  full_name='ga4gh.LinearAlignment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='position', full_name='ga4gh.LinearAlignment.position', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='mapping_quality', full_name='ga4gh.LinearAlignment.mapping_quality', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='cigar', full_name='ga4gh.LinearAlignment.cigar', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=791,
  serialized_end=901,
)


_READALIGNMENT_INFOENTRY = _descriptor.Descriptor(
  name='InfoEntry',
  full_name='ga4gh.ReadAlignment.InfoEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ga4gh.ReadAlignment.InfoEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='ga4gh.ReadAlignment.InfoEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=583,
  serialized_end=654,
)

_READALIGNMENT = _descriptor.Descriptor(
  name='ReadAlignment',
  full_name='ga4gh.ReadAlignment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='ga4gh.ReadAlignment.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='read_group_id', full_name='ga4gh.ReadAlignment.read_group_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fragment_name', full_name='ga4gh.ReadAlignment.fragment_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='improper_placement', full_name='ga4gh.ReadAlignment.improper_placement', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='duplicate_fragment', full_name='ga4gh.ReadAlignment.duplicate_fragment', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='number_reads', full_name='ga4gh.ReadAlignment.number_reads', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fragment_length', full_name='ga4gh.ReadAlignment.fragment_length', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='read_number', full_name='ga4gh.ReadAlignment.read_number', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='failed_vendor_quality_checks', full_name='ga4gh.ReadAlignment.failed_vendor_quality_checks', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='alignment', full_name='ga4gh.ReadAlignment.alignment', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='secondary_alignment', full_name='ga4gh.ReadAlignment.secondary_alignment', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='supplementary_alignment', full_name='ga4gh.ReadAlignment.supplementary_alignment', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aligned_sequence', full_name='ga4gh.ReadAlignment.aligned_sequence', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aligned_quality', full_name='ga4gh.ReadAlignment.aligned_quality', index=13,
      number=14, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='next_mate_position', full_name='ga4gh.ReadAlignment.next_mate_position', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='info', full_name='ga4gh.ReadAlignment.info', index=15,
      number=16, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_READALIGNMENT_INFOENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=904,
  serialized_end=1459,
)


_CIGARUNIT = _descriptor.Descriptor(
  name='CigarUnit',
  full_name='ga4gh.CigarUnit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='operation', full_name='ga4gh.CigarUnit.operation', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='operation_length', full_name='ga4gh.CigarUnit.operation_length', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='reference_sequence', full_name='ga4gh.CigarUnit.reference_sequence', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CIGARUNIT_OPERATION,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1462,
  serialized_end=1752,
)

_READGROUP_INFOENTRY.fields_by_name['value'].message_type = google_dot_protobuf_dot_struct__pb2._LISTVALUE
_READGROUP_INFOENTRY.containing_type = _READGROUP
_READGROUP.fields_by_name['experiment'].message_type = ga4gh_dot_assay__metadata__pb2._EXPERIMENT
_READGROUP.fields_by_name['stats'].message_type = _READSTATS
_READGROUP.fields_by_name['programs'].message_type = ga4gh_dot_metadata__pb2._PROGRAM
_READGROUP.fields_by_name['info'].message_type = _READGROUP_INFOENTRY
_READGROUPSET.fields_by_name['stats'].message_type = _READSTATS
_READGROUPSET.fields_by_name['read_groups'].message_type = _READGROUP
_LINEARALIGNMENT.fields_by_name['position'].message_type = ga4gh_dot_common__pb2._POSITION
_LINEARALIGNMENT.fields_by_name['cigar'].message_type = _CIGARUNIT
_READALIGNMENT_INFOENTRY.fields_by_name['value'].message_type = google_dot_protobuf_dot_struct__pb2._LISTVALUE
_READALIGNMENT_INFOENTRY.containing_type = _READALIGNMENT
_READALIGNMENT.fields_by_name['alignment'].message_type = _LINEARALIGNMENT
_READALIGNMENT.fields_by_name['next_mate_position'].message_type = ga4gh_dot_common__pb2._POSITION
_READALIGNMENT.fields_by_name['info'].message_type = _READALIGNMENT_INFOENTRY
_CIGARUNIT.fields_by_name['operation'].enum_type = _CIGARUNIT_OPERATION
_CIGARUNIT_OPERATION.containing_type = _CIGARUNIT
DESCRIPTOR.message_types_by_name['ReadStats'] = _READSTATS
DESCRIPTOR.message_types_by_name['ReadGroup'] = _READGROUP
DESCRIPTOR.message_types_by_name['ReadGroupSet'] = _READGROUPSET
DESCRIPTOR.message_types_by_name['LinearAlignment'] = _LINEARALIGNMENT
DESCRIPTOR.message_types_by_name['ReadAlignment'] = _READALIGNMENT
DESCRIPTOR.message_types_by_name['CigarUnit'] = _CIGARUNIT

ReadStats = _reflection.GeneratedProtocolMessageType('ReadStats', (_message.Message,), dict(
  DESCRIPTOR = _READSTATS,
  __module__ = 'ga4gh.reads_pb2'
  # @@protoc_insertion_point(class_scope:ga4gh.ReadStats)
  ))
_sym_db.RegisterMessage(ReadStats)

ReadGroup = _reflection.GeneratedProtocolMessageType('ReadGroup', (_message.Message,), dict(

  InfoEntry = _reflection.GeneratedProtocolMessageType('InfoEntry', (_message.Message,), dict(
    DESCRIPTOR = _READGROUP_INFOENTRY,
    __module__ = 'ga4gh.reads_pb2'
    # @@protoc_insertion_point(class_scope:ga4gh.ReadGroup.InfoEntry)
    ))
  ,
  DESCRIPTOR = _READGROUP,
  __module__ = 'ga4gh.reads_pb2'
  # @@protoc_insertion_point(class_scope:ga4gh.ReadGroup)
  ))
_sym_db.RegisterMessage(ReadGroup)
_sym_db.RegisterMessage(ReadGroup.InfoEntry)

ReadGroupSet = _reflection.GeneratedProtocolMessageType('ReadGroupSet', (_message.Message,), dict(
  DESCRIPTOR = _READGROUPSET,
  __module__ = 'ga4gh.reads_pb2'
  # @@protoc_insertion_point(class_scope:ga4gh.ReadGroupSet)
  ))
_sym_db.RegisterMessage(ReadGroupSet)

LinearAlignment = _reflection.GeneratedProtocolMessageType('LinearAlignment', (_message.Message,), dict(
  DESCRIPTOR = _LINEARALIGNMENT,
  __module__ = 'ga4gh.reads_pb2'
  # @@protoc_insertion_point(class_scope:ga4gh.LinearAlignment)
  ))
_sym_db.RegisterMessage(LinearAlignment)

ReadAlignment = _reflection.GeneratedProtocolMessageType('ReadAlignment', (_message.Message,), dict(

  InfoEntry = _reflection.GeneratedProtocolMessageType('InfoEntry', (_message.Message,), dict(
    DESCRIPTOR = _READALIGNMENT_INFOENTRY,
    __module__ = 'ga4gh.reads_pb2'
    # @@protoc_insertion_point(class_scope:ga4gh.ReadAlignment.InfoEntry)
    ))
  ,
  DESCRIPTOR = _READALIGNMENT,
  __module__ = 'ga4gh.reads_pb2'
  # @@protoc_insertion_point(class_scope:ga4gh.ReadAlignment)
  ))
_sym_db.RegisterMessage(ReadAlignment)
_sym_db.RegisterMessage(ReadAlignment.InfoEntry)

CigarUnit = _reflection.GeneratedProtocolMessageType('CigarUnit', (_message.Message,), dict(
  DESCRIPTOR = _CIGARUNIT,
  __module__ = 'ga4gh.reads_pb2'
  # @@protoc_insertion_point(class_scope:ga4gh.CigarUnit)
  ))
_sym_db.RegisterMessage(CigarUnit)


_READGROUP_INFOENTRY.has_options = True
_READGROUP_INFOENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
_READALIGNMENT_INFOENTRY.has_options = True
_READALIGNMENT_INFOENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
# @@protoc_insertion_point(module_scope)



"""
Generate a random FASTA file
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import hashlib
import json
import math
import os
import random

import utils


class FastaGenerator(object):
    """
    Generates a random FASTA file and metadata json.
    """
    def __init__(self, args):
        self.numBases = args.num_bases
        self.outputPrefix = args.output_prefix
        self.fastaFileName = "{}.fa".format(self.outputPrefix)
        self.referenceId = os.path.split(args.output_prefix)[-1]
        self.bases = ""

    def writeFasta(self):
        """
        Write the random fasta file
        """
        utils.log("writing {} bases to {} ...".format(
            self.numBases, self.fastaFileName))
        with open(self.fastaFileName, 'w') as fastaFile:
            firstLine = ">{} Generated by generate_fasta.py".format(
                self.referenceId)
            print(firstLine, file=fastaFile)
            basesPerLine = 70
            numLines = int(math.ceil(self.numBases / basesPerLine))
            baseChoices = ['A', 'G', 'C', 'T']
            basesRemaining = self.numBases
            for i in range(numLines):
                if basesRemaining < basesPerLine:
                    basesToWrite = basesRemaining
                else:
                    basesToWrite = basesPerLine
                bases = ''.join(
                    [random.choice(baseChoices) for _ in range(basesToWrite)])
                line = "{}".format(bases)
                self.bases += line
                print(line, file=fastaFile)
                basesRemaining -= basesToWrite
            assert basesRemaining == 0

    def writeMetadata(self):
        """
        Write some metadata.
        """
        metadata = {
            "md5checksum": hashlib.md5(self.bases).hexdigest(),
            "sourceUri": "http://example.com/random_url",
            "ncbiTaxonId": random.randint(1, 10000),
            "isDerived": False,
            "sourceDivergence": None,
            "sourceAccessions": [],
        }
        jsonFileName = "{}.json".format(self.outputPrefix)
        utils.log("writing metadata to {} ...".format(jsonFileName))
        with open(jsonFileName, "w") as jsonFile:
            json.dump(metadata, jsonFile, indent=4)

    def zipFasta(self):
        """
        Compress the fasta file
        """
        utils.log("zipping {} ...".format(self.fastaFileName))
        cmd = "bgzip -f {}".format(self.fastaFileName)
        utils.runCommand(cmd)

    def indexFasta(self):
        """
        Create index on the fasta file
        """
        zipFileName = "{}.gz".format(self.fastaFileName)
        utils.log("indexing {} ...".format(zipFileName))
        cmd = "samtools faidx {}".format(zipFileName)
        utils.runCommand(cmd)

    def generate(self):
        self.writeFasta()
        self.writeMetadata()
        self.zipFasta()
        self.indexFasta()


@utils.Timed()
def main():
    requiredExecutables = ['bgzip', 'samtools']
    utils.requireExecutables(requiredExecutables)
    parser = argparse.ArgumentParser(
        description="Generate random FASTA files and metadata")
    parser.add_argument(
        "output_prefix", help="The prefix for generated files.")
    basesDefault = 1000
    parser.add_argument(
        "--num-bases", "-n", default=basesDefault,
        help="number of bases to include; default {}".format(basesDefault))
    fastaGenerator = FastaGenerator(parser.parse_args())
    fastaGenerator.generate()


if __name__ == '__main__':
    main()
